import gzip
import os

from datetime import datetime
from pathlib import Path
from typing import Dict, List

from numpy import int64, nan
from pandas import DataFrame

from .exceptions import FilenameNotSetError, DataFrameNotCreatedError
from .rstndownloader import RSTNDownloader
from .rstnfile import RSTNFile


class RSTN:
    """Read RSTN files.

    Attributes
    ----------
    year: int
        Event's year.
    month: int
        Event's month.
    day: int
        Event's day.
    path: Path
        Where the files are/will be stored.
    station: str
        Station.
    dataframe: pandas.DataFrame
        The dataframe containing the data. (default: None)

    """

    def __init__(self, year: int, month: int, day: int, path: str,
                 station: str) -> None:
        self.year = str(year)
        self.month = self.__format_month(month)
        self.day = self.__format_day(day)

        self.path = self.__validate_path(path)
        self.__station = station
        self.__file: RSTNFile = RSTNFile(self.year, self.month, self.day,
                                         self.station)

        self.downloader = RSTNDownloader(self.year, self.month, self.day,
                                         self.path, station)
        self.dataframe = None
        self.frequencies_columns = [
            "245", "410", "610", "1415",
            "2695", "4995", "8800", "15400"
        ]

    @property
    def filename(self) -> str:
        """Gets the filename.

        Returns
        -------
        str
            Filename

        """

        return self.__file.name

    @property
    def station(self) -> str:
        """Gets the station name.

        Returns
        -------
        str
            Station name.

        """

        return self.__station

    def __format_day(self, day: int) -> str:
        """Formats the day as a string in the format dd.

        Parameters
        ----------
        day: int
            The day to be formatted.

        Returns
        -------
        day: str
            Formatted day.

        """

        formatted_day = str(day)
        if len(formatted_day) == 1:
            formatted_day = '0' + formatted_day

        return formatted_day

    def __format_month(self, month: int) -> str:
        """Formats the month as a string in the format mm.

        Parameters
        ----------
        month: int
            The month to be formatted.

        Returns
        -------
        month: str
            Formatted month.

        """

        formatted_month = str(month)
        if len(formatted_month) == 1:
            formatted_month = '0' + formatted_month

        return formatted_month

    def __validate_path(self, path: str) -> Path:
        """Validates the existence of a given path.

        Parameters
        ----------
        path: str
            The original path.

        Returns
        -------
        validated_path: Path
            The validated path.

        """
        validated_path = Path(path)

        if not validated_path.exists():
            validated_path.mkdir()

        return validated_path

    def __cast_to_int64(self, number):
        """Casts a number to the numpy int64 type.

        Parameters
        ----------
        number: int or str
            The number that will be changed.

        Returns
        -------
        number: numpy.int64 or numpy.nan
            The number as a int64 or NaN.

        """

        try:
            number = int64(number)
        except ValueError:
            number = nan

        return number

    @staticmethod
    def is_gzippep(filename: Path) -> bool:
        try:
            gzip.open(filename, 'rb').read()
            return True
        except OSError:
            return False

    def decompress_file(self, filename: str) -> str:
        """Gets gzipped file content.

        The method doesn't actually decompress the file. In reallity it reads
        the compressed data and writes it in a new file without
        the .gz extension.

        Returns
        -------
        str
            Final filename.

        """

        path_to_gzip = self.path.joinpath(filename)

        if not self.is_gzippep(path_to_gzip):
            self.__file.name = filename
            return filename

        with gzip.open(path_to_gzip, 'rb') as gzipped_file:
            file_content = gzipped_file.read()
            final_name = filename.split('.gz')[0]
            with Path(self.path.joinpath(final_name)).open("wb") as final_file:
                final_file.write(file_content)

        os.remove(Path(self.path.joinpath(filename)))

        self.__file.name = final_name

        return final_name

    def __set_column_interval(self) -> int:
        """Sets the interval for each frequency column.

        The column size changes from time to time.

        Returns
        -------
        interval: int
            The column interval size.

        """

        if int(self.year) >= 2015:
            return 8
        elif int(self.year) >= 2008:
            return 7
        else:
            return 6

    def read_file(self) -> Dict[str, List]:
        """Reads the file data and saves it in columns by frequency.

        The first column is 18 digits long, the first 4 indicating the station,
        the others the timestamp. Each column for the frequency is 6 digits
        long if the year before 2007, if the year is 2008 or later than the
        column is 7 digits long.

        Returns
        -------
        dict
            The data in arrays.

        Raises
        ------
        FilenameNotSetError
            If filename is not set.

        """

        if self.__file is None:
            raise FilenameNotSetError(
                "The file " + self.__file.name + " has an invalid name.")

        rstn_data: Dict[str, List] = {
            "time": [], "245": [], "410": [], "610": [],
            "1415": [], "2695": [], "4995": [], "8800": [],
            "15400": []
        }

        interval = self.__set_column_interval()

        with Path(self.path.joinpath(self.__file.name)).open("r") as fp:
            for line in fp.readlines():
                year = int(line[4:8])
                month = int(line[8:10])
                day = int(line[10:12])
                hour = int(line[12:14])
                minute = int(line[14:16])
                second = int(line[16:18])

                date = datetime(year, month, day, hour, minute, second)

                rstn_data["time"].append(date)
                rstn_data["245"].append(self.__cast_to_int64(
                    line[18:18+interval]))
                rstn_data["410"].append(self.__cast_to_int64(
                    line[18+interval:18+interval*2]))
                rstn_data["610"].append(self.__cast_to_int64(
                    line[18+interval*2:18+interval*3]))
                rstn_data["1415"].append(self.__cast_to_int64(
                    line[18+interval*3:18+interval*4]))
                rstn_data["2695"].append(self.__cast_to_int64(
                    line[18+interval*4:18+interval*5]))
                rstn_data["4995"].append(self.__cast_to_int64(
                    line[18+interval*5:18+interval*6]))
                rstn_data["8800"].append(self.__cast_to_int64(
                    line[18+interval*6:18+interval*7]))
                rstn_data["15400"].append(self.__cast_to_int64(
                    line[18+interval*7:]))

        return rstn_data

    def create_dataframe(self):
        """Creates the dataframe with the file's data.

        Returns
        -------
        pandas.DataFrame
            The dataframe with the data.

        Raises
        ------
        FileNotFoundError
            If the filename is not set, therefore, it can't be found.

        """

        try:
            data = self.read_file()
        except FilenameNotSetError:
            raise FileNotFoundError(
                "The file " + self.__file.name + " was not found.")

        self.dataframe = DataFrame(data, columns=self.frequencies_columns,
                                   index=data["time"])

        self.dataframe.index = self.dataframe.index.tz_localize("UTC")

        return self.dataframe

    def plot(self):
        """Plots the file's data.

        Returns
        -------
        axis: matplotlib.Axes
            Graphic's axes object for manipulation.

        Raises
        ------
        DataFrameNotCreatedError:
            If the dataframe was not created before this method is called.

        """

        if self.dataframe is None:
            raise DataFrameNotCreatedError(
                "Can't plot data. Dataframe was not created.")

        axis = self.dataframe.plot()

        return axis



from pathlib import Path
from typing import Union
from urllib.error import HTTPError

from requests import get

from .exceptions import FileNotFoundOnServerError, InvalidDateError
from .rstnfile import RSTNFile


class RSTNDownloader:
    """Download rstn 1 second data from NOAA's site.

    Base url to NGDC data is: https://www.ngdc.noaa.gov/

    Rstn 1 second data is under Solar-Terrestrial Physics(STP):
    stp/space-weather/solar-data/solar-features/solar-radio/rstn-1-second/

    Attributes
    ----------
    year: int or str
        Event's year.
    month: int or str
        Event's month.
    day: int or str
        Event's day.
    path: Path
        Where the files are/will be stored.
    __file: RSTNFile
        File object used to manipulate the file name properties.
    __base_uri: str
        The base uri to noaa's.

    """

    def __init__(self, year: str, month: str, day: str, path: Path,
                 station: str) -> None:
        self.year = year
        self.month = month
        self.day = day
        self.path = path
        self.__file = RSTNFile(year, month, day, station)
        self.__base_uri = "https://www.ngdc.noaa.gov"

    def file_exists(self) -> Union[bool, str]:
        """Checks if the file exists.

        Returns
        -------
        bool
            If the file exists.

        """

        filename_upper = self.__file.set_filename(
            True) + self.__file.set_file_extension_upper(False)
        filename_lower = self.__file.set_filename(
            False) + self.__file.set_file_extension_lower(False)

        if Path(self.path.joinpath(filename_upper)).exists():
            return filename_upper

        if Path(self.path.joinpath(filename_lower)).exists():
            return filename_lower

        return False

    def __set_url(self, upper: bool) -> str:
        """Creates the url of the file to be downloaded.

        Parameters
        ----------
        upper: bool
            Used to try downloading both name in upper and lower case.

        Returns
        -------
        str
            The whole url.

        """

        station_name = self.__file.format_station_for_url()

        if upper:
            filename = self.__file.set_filename(True)
            file_extension = self.__file.set_file_extension_upper()
        else:
            filename = self.__file.set_filename(False)
            file_extension = self.__file.set_file_extension_lower()

        url = self.__base_uri + "/stp/space-weather/solar-data/"
        url += "solar-features/solar-radio/rstn-1-second/"
        url += station_name + '/' + self.year + '/' + self.month + '/'
        url += filename + file_extension

        return url

    def __download(self, upper: bool) -> str:
        """Downloads the gzipped file.

        Returns
        -------
        filename: str
            the saved filename.

        Raises
        ------
        HttpError
            Raised if the status code of the response is not 200.

        """

        url = self.__set_url(upper)
        filename = self.__file.set_filename(upper)

        if upper:
            filename += self.__file.set_file_extension_upper()
        else:
            filename += self.__file.set_file_extension_lower()

        r = get(url, allow_redirects=False)
        if r.status_code != 200:
            raise HTTPError(url, r.status_code, "HTTP error", r.headers, "")

        with open(filename, 'wb') as fp:
            fp.write(r.content)

        return filename

    def download_file(self) -> str:
        """Downloads the file via https.

        Tries to download with the file extension in upper case.
        Then tries to download with the file extension in lower case.

        Returns
        -------
        filename: str
            The downloaded gzipped filename.

        Raises
        ------
        InvalidDateError
            If the date is not valid

        FileNotFoundOnServerError
            If the file does not exist in noaa's website.

        """

        if not self.__file.is_date_valid():
            raise InvalidDateError("Invalid date")

        if self.file_exists():
            return self.file_exists()

        try:
            filename = self.__download(upper=True)
        except HTTPError:
            try:
                filename = self.__download(upper=False)
            except HTTPError:
                url = self.__set_url(upper=False)
                raise FileNotFoundOnServerError(
                    "The file on: " + url + " was not found on server.")

        Path(filename).rename(self.path.joinpath(filename))
        return filename



from datetime import datetime


class RSTNFile:
    """RSTN filename manipulation tools.

    Attributes
    ----------
    day: str
        Event's day.
    month: str
        Event's month.
    year: str
        Event's year.
    station: str
        Station.
    name: str
        Filename
    __station_extensions: Dict
        All possible file extensions for each station.

    """

    def __init__(self, year: str, month: str, day: str, station: str) -> None:
        self.year = year
        self.month = month
        self.day = day
        self.station = station
        self.name = self.__set_default_name()
        self.__station_extensions = {
            "sagamore hill": {
                "lower": "k7o", "upper": "K7O"
            },
            "san vito": {
                "lower": "lis", "upper": "LIS"
            },
            "palehua": {
                "lower": "phf", "upper": "PHF"
            },
            "learmonth": {
                "lower": "apl", "upper": "APL"
            }
        }

    def __change_month_upper(self) -> str:
        """Sets the month for the filename in upper case.

        Returns
        -------
        str
            The month in upper case.

        """

        months = [
            "JAN", "FEB", "MAR", "APR", "MAY", "JUN",
            "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"
        ]

        # Returns the corresponding month to download the file.
        index = int(self.month) - 1
        return months[index]

    def __change_month_lower(self) -> str:
        """Sets the month for the filename in lowercase.

        Returns
        -------
        str
            The month in lower case.

        """

        months = [
            "jan", "feb", "mar", "apr", "may", "jun",
            "jul", "aug", "sep", "oct", "nov", "dec"
        ]

        # Returns the corresponding month to download the file.
        index = int(self.month) - 1
        return months[index]

    def set_file_extension_upper(self, file_gzip: bool = True) -> str:
        """Creates the file extension upper case.

        Parameters
        ----------
        file_gzip: bool
            Sets if the extension will have .gz.

        Returns
        -------
        str
            The file extension.

        """

        extension = "." + \
            self.__station_extensions[self.station.lower()]["upper"]

        if file_gzip:
            return extension + ".gz"

        return extension

    def set_file_extension_lower(self, file_gzip: bool = True) -> str:
        """Creates the file extension lower case.


        Parameters
        ----------
        file_gzip: bool
            Sets if the extension will have .gz.

        Returns
        -------
        str
            The file extension.

        """

        extension = "." + \
            self.__station_extensions[self.station.lower()]["lower"]

        if file_gzip:
            return extension + ".gz"

        return extension

    def set_filename(self, upper: bool) -> str:
        """Creates the filename.

        Parameters
        ----------
        upper: bool
            Sets if the filename is going to be upper or lower case.

        Returns
        -------
        str
            The filename.

        """

        if upper:
            filename = self.day + self.__change_month_upper() + self.year[2:]
        else:
            filename = self.day + self.__change_month_lower() + self.year[2:]

        return filename

    def __set_default_name(self) -> str:
        """Sets the default filename.

        Returns
        -------
        str
            The filename

        """
        return self.set_filename(True)

    def format_station_for_url(self) -> str:
        """Formats the station name as it is in NOAA's site for the url.

        Returns
        -------
        str
            The station name as it is in the site url.

        """

        formatted_station = self.station.lower().replace(' ', '-')

        return formatted_station

    def is_date_valid(self):
        """Checks if the date is valid.

        Returns
        -------
        bool
            If the date is valid.

        """

        try:
            year = int(self.year)
            month = int(self.month)
            day = int(self.day)
            datetime(year, month, day)
            return True
        except ValueError:
            return False



from setuptools import setup

with open("README.org", "r") as readme:
    long_description = readme.read()

setup(
    name="rstnpy",
    version="1.0",
    author="Edison Neto",
    author_email="ednetoali@gmail.com",
    description="Library used to work with rstn-1-second data",
    long_description=long_description,
    url="https://github.com/craam/rstnpy",
    packages=['rstnpy'],
    include_package_data=True,
    install_requires=[
        'requests>=2.20',
        'numpy>=1.13',
        'pandas>=0.21'
        'matplotlib>=2.2.2'
    ],
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
)
